local RUF = LibStub("AceAddon-3.0"):GetAddon("RUF")
local L = LibStub("AceLocale-3.0"):GetLocale("RUF")
local _, ns = ...
local oUF = ns.oUF
local cfg = RUF_cfg
local tags = oUF.Tags.Methods or oUF.Tags
local events = oUF.TagEvents or oUF.Tags.Events

local function Short(value,format)
	if type(value) == "number" then
		local fmt
		if value >= 1000000000 or value <= -1000000000 then
			fmt = "%.1fB"
			value = value / 1000000000
		elseif value >= 10000000 or value <= -10000000 then
			fmt = "%.1fM"
			value = value / 1000000
		elseif value >= 1000000 or value <= -1000000 then
			fmt = "%.2fM"
			value = value / 1000000
		elseif value >= 100000 or value <= -100000 then
			fmt = "%.0fK"
			value = value / 1000
		elseif value >= 10000 or value <= -10000 then
			fmt = "%.1fK"
			value = value / 1000
		else
			fmt = "%d"
			value = math.floor(value + 0.5)
		end
		if format then
			return fmt:format(value)
		end
		return fmt, value
	else
		local fmt_a, fmt_b
		local a, b = value:match("^(%d+)/(%d+)$")
		if a then
			a, b = tonumber(a), tonumber(b)
			if a >= 1000000000 or a <= -1000000000 then
				fmt_a = "%.1fB"
				a = a / 1000000000
			elseif a >= 10000000 or a <= -10000000 then
				fmt_a = "%.1fM"
				a = a / 1000000
			elseif a >= 1000000 or a <= -1000000 then
				fmt_a = "%.2fM"
				a = a / 1000000
			elseif a >= 100000 or a <= -100000 then
				fmt_a = "%.0fK"
				a = a / 1000
			elseif a >= 10000 or a <= -10000 then
				fmt_a = "%.1fK"
				a = a / 1000
			end
			if b >= 1000000000 or b <= -1000000000 then
				fmt_b = "%.1fB"
				b = b / 1000000000
			elseif b >= 10000000 or b <= -10000000 then
				fmt_b = "%.1fM"
				b = b / 1000000
			elseif b >= 1000000 or b <= -1000000 then
				fmt_b = "%.2fM"
				b = b / 1000000
			elseif b >= 100000 or b <= -100000 then
				fmt_b = "%.0fK"
				b = b / 1000
			elseif b >= 10000 or b <= -10000 then
				fmt_b = "%.1fK"
				b = b / 1000
			end
			local fmt = ("%s/%s"):format(fmt_a, fmt_b)
			if format then
				return fmt:format(a, b)
			end
			return fmt, a, b
		else
			return value
		end
	end
end

local function Round(number, digits)
	local mantissa = 10^(digits or 0)
	local norm = number * mantissa + 0.5
	local norm_floor = math.floor(norm)
	if norm == norm_floor and (norm_floor % 2) == 1 then
		return (norm_floor - 1) / mantissa
	end
	return norm_floor / mantissa
end

local function Percent(x, y)
	if y ~= 0 then
		return Round(x / y * 100, 1)
	end
	return 0
end

local function colorsAndPercent(a, b, ...)
	if(a <= 0 or b == 0) then
		return nil, ...
	elseif(a >= b) then
		return nil, select(select('#', ...) - 2, ...)
	end

	local num = select('#', ...) / 3
	local segment, relperc = math.modf((a / b) * (num - 1))
	return relperc, select((segment * 3) + 1, ...)
end

local function ColorGradient(...)
	local relperc, r1, g1, b1, r2, g2, b2 = colorsAndPercent(...)
	if(relperc) then
		return r1 + (r2 - r1) * relperc, g1 + (g2 - g1) * relperc, b1 + (b2 - b1) * relperc
	else
		return r1, g1, b1
	end
end


local function ReturnTextColors(unit, tag, cur, max) -- Get Text Colors
	--local cur, max = UnitHealth(unit), UnitHealthMax(unit)
	local r,g,b = unpack(RUF.db.profile.Font[tag].Color)
	local class = select(2,UnitClass(unit))
	if not class then class = "PRIEST" end
	if RUF.db.profile.Font[tag].colorGradient and RUF.db.profile.Font[tag].colorGradientAtMax and cur == max then -- If we want to show gradient colors at max health, and we're at max health.
		if tag == "Power" then -- Return Power Gradient Colors
			r,g,b = ColorGradient(cur, max, unpack(RUF.db.profile.Colors.Power.TextGradient))
		else -- Return Health Gradient Colors
			r,g,b = ColorGradient(cur, max, unpack(RUF.db.profile.Colors.Health.TextGradient))
		end
	elseif RUF.db.profile.Font[tag].colorGradient and cur < max and cur > 0 then -- If we want to show gradient colors and we're not at max health.
		if tag == "Power" then -- Return Power Gradient Colors
			r,g,b = ColorGradient(cur, max, unpack(RUF.db.profile.Colors.Power.TextGradient))
		else -- Return Health Gradient Colors
			r,g,b = ColorGradient(cur, max, unpack(RUF.db.profile.Colors.Health.TextGradient))
		end
	elseif RUF.db.profile.Font[tag].colorClass and UnitIsPlayer(unit) then -- If we want to show class colors.
		r,g,b = unpack(RUF.db.profile.Colors.ClassColors[class])
	elseif RUF.db.profile.Font[tag].colorReaction then -- If we want to show unit reaction colors
		if UnitPlayerControlled(unit) and not UnitCanAttack("target","player") and not UnitIsPlayer(unit) then  -- If the unit is an allied pet then show as blue.
			r,g,b = unpack(RUF.db.profile.Colors.ReactionColors[10])
		elseif UnitReaction(unit,'player') then -- If the unit is an offline party member, possibly others too?
			r,g,b = unpack(RUF.db.profile.Colors.ReactionColors[UnitReaction(unit, 'player')])
		elseif UnitInParty(unit) then
			r,g,b = unpack(RUF.db.profile.Colors.ReactionColors[5]) -- So Reaction Works when Party member is in a different zone and UnitReaction returns nil
		else
			r,g,b = unpack(RUF.db.profile.Font[tag].Color) 
		end
	elseif RUF.db.profile.Font[tag].colorPower then -- Color by UnitPower (Mana, Rage, etc.)
		local pType,pToken,altr,altg,altb = UnitPowerType(unit)
		r,g,b = unpack(RUF.db.profile.Colors.PowerColors[pType])	
	else -- If none of that matches, show the base colour
		r,g,b = unpack(RUF.db.profile.Font[tag].Color)
	end
	return r,g,b
end


-- Uppercase Name
-- Limits NPC names to 18 characters.
tags['Raeli:UpperName'] = function(unit)
	if not UnitName(unit) then return end

	-- Substitute Nickname
	local name
	if RUF.db.profile.Font.Name.Case == 1 then
		name = string.upper(UnitName(unit))
	elseif RUF.db.profile.Font.Name.Case == 2 then
		name = string.lower(UnitName(unit))
	else
		name = UnitName(unit)
	end
	local player = string.upper(UnitName("player"))
	if RUF.db.char.unit["player"].NickName and string.upper(name) == player and string.len(RUF.db.char.unit["player"].NickName) > 0 then
		if RUF.db.profile.Font.Name.Case == 1 then
			name = string.upper(RUF.db.char.unit["player"].NickName)
		elseif RUF.db.profile.Font.Name.Case == 2 then
			name = string.lower(RUF.db.char.unit["player"].NickName)
		else
			name = RUF.db.char.unit["player"].NickName
		end
	end

	local cur, max = UnitHealth(unit), UnitHealthMax(unit)
	local r,g,b = ReturnTextColors(unit, "Name", cur, max)
	-- Return Text String
	if UnitIsPlayer(unit) then
		return string.format('|cff%02x%02x%02x%s|r',r*255,g*255,b*255,name)
	else
		return string.format('|cff%02x%02x%02x%s|r',r*255,g*255,b*255,string.sub(name,1,18))
	end
end
events['Raeli:UpperName'] = 'UNIT_NAME_UPDATE'

-- Health
-- Class Coloured if Player, Reaction Coloured if NPC, Gradient Coloured if below 100%
tags['Raeli:FullHealth'] = function(unit)
	if not UnitName(unit) then return end
	local cur, max = UnitHealth(unit), UnitHealthMax(unit)
	local r,g,b = ReturnTextColors(unit, "Health", cur, max)
	
	-- Return Text String
	if UnitIsDead(unit) then
		if RUF.db.profile.Font.Health.Case == 1 then
			return string.format('|cff%02x%02x%02x%s|r',r*255,g*255,b*255,string.upper(L["Dead"]))
		elseif RUF.db.profile.Font.Health.Case == 2 then
			return string.format('|cff%02x%02x%02x%s|r',r*255,g*255,b*255,string.lower(L["Dead"]))
		else
			return string.format('|cff%02x%02x%02x%s|r',r*255,g*255,b*255,L["Dead"])
		end		
	elseif UnitIsGhost(unit) then
		if RUF.db.profile.Font.Health.Case == 1 then
			return string.format('|cff%02x%02x%02x%s|r',r*255,g*255,b*255,string.upper(L["Ghost"]))
		elseif RUF.db.profile.Font.Health.Case == 2 then
			return string.format('|cff%02x%02x%02x%s|r',r*255,g*255,b*255,string.lower(L["Ghost"]))
		else
			return string.format('|cff%02x%02x%02x%s|r',r*255,g*255,b*255,L["Ghost"])
		end
	elseif not UnitIsConnected(unit) then
		if RUF.db.profile.Font.Health.Case == 1 then
			return string.format('|cff%02x%02x%02x%s|r',r*255,g*255,b*255,string.upper(L["Offline"]))
		elseif RUF.db.profile.Font.Health.Case == 2 then
			return string.format('|cff%02x%02x%02x%s|r',r*255,g*255,b*255,string.lower(L["Offline"]))
		else
			return string.format('|cff%02x%02x%02x%s|r',r*255,g*255,b*255,L["Offline"])
		end
	elseif cur == max then -- if we're at full health 
		if RUF.db.profile.Font.Health.Case == 1 then
			return string.format('|cff%02x%02x%02x%s|r',r*255,g*255,b*255,Short(cur,true))
		elseif RUF.db.profile.Font.Health.Case == 2 then
			return string.format('|cff%02x%02x%02x%s|r',r*255,g*255,b*255,string.lower(Short(cur,true)))
		else
			return string.format('|cff%02x%02x%02x%s|r',r*255,g*255,b*255,Short(cur,true))
		end		
	else
		if RUF.db.profile.Font.Health.Case == 1 then
			return string.format('|cff%02x%02x%02x%s%s%s%%|r',r*255,g*255,b*255,Short(cur,true), " - ",Percent(cur,max))
		elseif RUF.db.profile.Font.Health.Case == 2 then
			return string.format('|cff%02x%02x%02x%s%s%s%%|r',r*255,g*255,b*255,(Short(cur,true)), " - ",Percent(cur,max))
		else
			return string.format('|cff%02x%02x%02x%s%s%s%%|r',r*255,g*255,b*255,string.lower(Short(cur,true)), " - ",Percent(cur,max))
		end		
	end
end
events['Raeli:FullHealth'] = 'UNIT_HEALTH UNIT_CONNECTION'

-- Power value
-- Percentage shown if below 100%, hidden if 0
tags['Raeli:Power'] = function(unit)
	local cur, max = UnitPower(unit), UnitPowerMax(unit)
	local pType,pToken,altr,altg,altb = UnitPowerType(unit)
	--if (altr ~= nil) or (altg ~= nil) or (altb ~= nil) then
	--	local pr,pg,pb = altr or 0,altg or 0,altb or 0
	--else
		local cur, max = UnitPower(unit), UnitPowerMax(unit)
		local r,g,b = ReturnTextColors(unit, "Power", cur, max)
		--local pr,pg,pb = unpack(RUF.db.profile.Colors.PowerColors[pType])
	--end


	if RUF.db.profile.unit[unit].Texts.Power == 0 then -- Always Hidden on this unit.
		return ''
	elseif RUF.db.profile.unit[unit].Texts.Power == 1 then -- Hidden at 0
		if max < 1 then
			return ''
		elseif cur > 0 and cur < max then
			if RUF.db.profile.Font.Power.Case == 2 then
				return string.format('|cff%02x%02x%02x%s%s%s%%|r',r*255,g*255,b*255,string.lower(Short(cur,true)), " - ",Percent(cur,max))
			else
				return string.format('|cff%02x%02x%02x%s%s%s%%|r',r*255,g*255,b*255,Short(cur,true), " - ",Percent(cur,max))
			end	
		elseif cur == max then
			if RUF.db.profile.Font.Power.Case == 2 then
				return string.format('|cff%02x%02x%02x%s|r',r*255,g*255,b*255,string.lower(Short(cur,true)))
			else
				return string.format('|cff%02x%02x%02x%s|r',r*255,g*255,b*255,Short(cur,true))
			end
		else
			return ''			
		end
	elseif 	RUF.db.profile.unit[unit].Texts.Power == 2 then -- Always Show
		if cur == max then
			if RUF.db.profile.Font.Power.Case == 2 then
				return string.format('|cff%02x%02x%02x%s|r',r*255,g*255,b*255,string.lower(Short(cur,true)))
			else
				return string.format('|cff%02x%02x%02x%s|r',r*255,g*255,b*255,Short(cur,true))
			end
		elseif cur < max and cur > 0 then
			if RUF.db.profile.Font.Power.Case == 2 then
				return string.format('|cff%02x%02x%02x%s%s%s%%|r',r*255,g*255,b*255,string.lower(Short(cur,true)), " - ",Percent(cur,max))
			else
				return string.format('|cff%02x%02x%02x%s%s%s%%|r',r*255,g*255,b*255,Short(cur,true), " - ",Percent(cur,max))
			end
		else
			if RUF.db.profile.Font.Power.Case == 2 then
				return string.format('|cff%02x%02x%02x%s|r',r*255,g*255,b*255,"0")
			else
				return string.format('|cff%02x%02x%02x%s|r',r*255,g*255,b*255,"0")
			end
		end
	end

	
	--[[if max == 0 then
		return ''
	elseif cur == max then
		return string.format('|cff%02x%02x%02x%s|r',r*255,g*255,b*255,Short(cur,true))
	elseif cur < max and cur > 0 then
		return string.format('|cff%02x%02x%02x%s%s%s%%|r',r*255,g*255,b*255,Short(cur,true), " - ",Percent(cur,max))
	else
		return ''
	end]]--

end
events['Raeli:Power'] = 'UNIT_POWER'


-- Unit smart level
tags['Raeli:Level'] = function(unit)
	local l = UnitLevel(unit)
	local c = GetQuestDifficultyColor(l)  
	if l <= 0 then l = "??" end  
	if l ~= UnitLevel("player") then
		return string.format('|cff%02x%02x%02x%s|r',c.r*255,c.g*255,c.b*255,l)
	else
		return ''
	end
end
events['Raeli:Level'] = 'UNIT_LEVEL PLAYER_LEVEL_UP UNIT_CLASSIFICATION_CHANGED'
oUF.Tags.SharedEvents.EVENT_NAME = true


tags['Raeli:AFKDND'] = function(unit)
	if UnitIsAFK(unit) then
		return string.format('|cffFFFFFF[AFK]|r')
	end
	if UnitIsDND(unit) then
		return string.format('|cffFFFFFF[DND]|r')
	end
end
events['Raeli:AFKDND'] = 'PLAYER_FLAGS_CHANGED'




--[[
if RUF.db.profile.unit[unit].Bars.Power == 0 and element.__owner.Power:IsShown() then -- Always Hidden on this unit.
	element.__owner.Power:Hide()
	element.__owner.Power.bg:Hide()
elseif RUF.db.profile.unit[unit].Bars.Power == 1 then -- Hidden at 0
	if cur < 1 and element.__owner.Power:IsShown() then
		element.__owner.Power:Hide()
		element.__owner.Power.bg:Hide()
	end
	if cur > 0 and not element.__owner.Power:IsShown() then 
		element.__owner.Power:Show()
		element.__owner.Power.bg:Show() 
	end
elseif 	RUF.db.profile.unit[unit].Bars.Power == 2 and not element.__owner.Power:IsShown() then		
	element.__owner.Power:Show()
	element.__owner.Power.bg:Show() 	
end]]--
